@@ -4,9 +4,11 @@ import os
 import traceback
 import re
 # Local modules
-from fortls.parse_fortran import process_file, read_use_stmt, detect_fixed_format, \
-    detect_comment_start
-from fortls.objects import find_in_scope, fortran_meth, get_use_tree
+from fortls.parse_fortran import process_file, read_use_stmt, read_var_def, \
+    detect_fixed_format, detect_comment_start
+from fortls.objects import find_in_scope, get_use_tree
+from fortls.intrinsics import get_keywords, get_intrinsics, get_intrinsic_modules, \
+    set_lowercase_intrinsics
 
 log = logging.getLogger(__name__)
 PY3K = sys.version_info >= (3, 0)
@@ -14,10 +16,16 @@ PY3K = sys.version_info >= (3, 0)
 FORTRAN_EXT_REGEX = re.compile(r'^\.F(77|90|95|03|08|OR|PP)?$', re.I)
 OBJBREAK_REGEX = re.compile(r'[\/\-(.,+*<>=$: ]', re.I)
 WORD_REGEX = re.compile(r'[a-z_][a-z0-9_]*', re.I)
-CALL_REGEX = re.compile(r'[ \t]*CALL[ \t]+[a-z0-9_%]*$', re.I)
-TYPE_STMNT_REGEX = re.compile(r'[ \t]*(TYPE|CLASS)[ \t]*(IS)?[ \t]*[a-z0-9_]*$', re.I)
+CALL_REGEX = re.compile(r'[ ]*CALL[ ]+[a-z0-9_%]*$', re.I)
+TYPE_STMNT_REGEX = re.compile(r'[ ]*(TYPE|CLASS)[ ]*(IS)?[ ]*$', re.I)
+TYPE_DEF_REGEX = re.compile(r'[ ]*TYPE[, ]+', re.I)
+EXTENDS_REGEX = re.compile(r'EXTENDS[ ]*$', re.I)
+PROCEDURE_STMNT_REGEX = re.compile(r'[ ]*(PROCEDURE)[ ]*$', re.I)
+SCOPE_DEF_REGEX = re.compile(r'[ ]*(MODULE|PROGRAM|SUBROUTINE|FUNCTION)[ ]+', re.I)
+END_REGEX = re.compile(r'[ ]*(END)( |MODULE|PROGRAM|SUBROUTINE|FUNCTION|TYPE|DO|IF|SELECT)?', re.I)
+IMPORT_REGEX = re.compile(r'[ ]*IMPORT[ ]+', re.I)
 FIXED_CONT_REGEX = re.compile(r'(     [\S])')
-FREE_OPT_CONT_REGEX = re.compile(r'([ \t]*&)')
+FREE_OPT_CONT_REGEX = re.compile(r'([ ]*&)')
 
 
 def path_from_uri(uri):
@@ -26,6 +34,7 @@ def path_from_uri(uri):
         return uri
     if os.name == "nt":
         _, path = uri.split("file:///", 1)
+        path = path.replace("%3A", ":")  # Replace escaped colon in Windows paths
     else:
         _, path = uri.split("file://", 1)
     return path
@@ -39,88 +48,91 @@ def path_to_uri(path):
         return "file://"+path
 
 
+def read_file_split(filepath):
+    # Read and add file from disk
+    try:
+        if PY3K:
+            with open(filepath, 'r', encoding="utf-8") as fhandle:
+                contents = re.sub(r'\t', r'  ', fhandle.read())
+                contents_split = contents.splitlines()
+        else:
+            with open(filepath, 'r') as fhandle:
+                contents = re.sub(r'\t', r'  ', fhandle.read())
+                contents_split = contents.splitlines()
+    except:
+        return None, 'Could not read/decode file'
+    else:
+        return contents_split, None
+
+
 def init_file(filepath):
     #
-    if PY3K:
-        with open(filepath, 'r', encoding="utf-8") as fhandle:
-            contents_split = fhandle.readlines()
-    else:
-        with open(filepath, 'r') as fhandle:
-            contents_split = fhandle.readlines()
+    contents_split, err_str = read_file_split(filepath)
+    if contents_split is None:
+        return None, err_str
     #
     try:
         fixed_flag = detect_fixed_format(contents_split)
         ast_new = process_file(contents_split, True, filepath, fixed_flag)
     except:
-        return None
-    else:
-        # Construct new file object and add to workspace
-        tmp_obj = {
-            "contents": contents_split,
-            "ast": ast_new,
-            "fixed": fixed_flag
-        }
-        return tmp_obj
+        return None, 'Error during parsing'
+    # Construct new file object and add to workspace
+    tmp_obj = {
+        "contents": contents_split,
+        "ast": ast_new,
+        "fixed": fixed_flag
+    }
+    return tmp_obj, None
 
 
 def tokenize_line(line):
-    var_list = []
-    paren_stack = [[]]
-    paren_count = [0]
-    iCurr = -1
-    for (i, character) in enumerate(line):
-        if character == "(":
-            iCurr += 1
-            if len(paren_stack) < iCurr+1:
-                paren_stack.append([])
-                paren_count.append(0)
-            paren_stack[iCurr].append([i, -1])
-        elif character == ")":
-            paren_stack[iCurr][-1][1] = i
-            iCurr -= 1
-    #
-    if len(paren_stack[0]) == 0:
-        var_list.append([[[[0, -1]], line]])
-    else:
-        var_groups = [[] for i in range(len(paren_stack)+1)]
-        var_list, out_len = paren_split(line, paren_stack, 0, var_groups)
-    return var_list
+    paren_list = [[[-1, len(line)]], []]
+    level = 1
+    in_string = False
+    string_char = ""
+    for i, char in enumerate(line):
+        if in_string:
+            if char == string_char:
+                in_string = False
+            continue
+        if char == '(':
+            paren_list[level].append([i, len(line)])
+            level += 1
+            if len(paren_list) < level+1:
+                paren_list.append([])
+        elif char == ')':
+            paren_list[level-1][-1][1] = i
+            level -= 1
+        elif char == "'" or char == '"':
+            in_string = True
+            string_char = char
+    return paren_split(line, paren_list[:-1])
 
 
-def paren_split(line, paren_groups, level, var_groups, i1=0, i2=-1):
-    tmp_list = []
-    tmp_str = ''
-    nline = len(line)
-    if i2 < 0:
-        i2 = nline
-    i0 = i1
-    for paren_group in paren_groups[level]:
-        pg0 = paren_group[0]
-        pg1 = paren_group[1]
-        if pg0 < i1 or pg1 > i2:
-            continue
-        tmp_str += line[i0:pg0]
-        tmp_list.append([i0, pg0])
-        if pg1 < 0:
-            pg1 = nline
-        if len(paren_groups) > level+1:
-            var_groups, out_len = \
-                paren_split(line, paren_groups, level+1,
-                            var_groups, pg0+1, pg1)
-            if out_len == 0:
-                var_groups[level+1].append([[[pg0+1, pg1]], line[pg0+1:pg1]])
-        else:
-            var_groups[level+1].append([[[pg0+1, pg1]], line[pg0+1:pg1]])
-        if pg1 == nline:
-            continue
-        i0 = pg1+1
-    #
-    if i0 > tmp_list[-1][1]:
-        tmp_str += line[i0:i2]
-        tmp_list.append([i0, i2])
-    out_list = [[tmp_list, tmp_str]]
-    var_groups[level].extend(out_list)
-    return var_groups, len(out_list)
+def paren_split(line, paren_list):
+    sections = []
+    for ilev, level in enumerate(paren_list):
+        sections.append([])
+        for igroup, group in enumerate(level):
+            i1 = group[0]
+            i2 = group[1]
+            tmp_str = ""
+            i3 = i1 + 1
+            ranges = []
+            if len(paren_list) > ilev+1:
+                for lower_group in paren_list[ilev+1]:
+                    if lower_group[0] > i1 and lower_group[1] <= i2:
+                        tmp_str += line[i3:lower_group[0]]
+                        ranges.append([i3, lower_group[0]])
+                        i3 = lower_group[1] + 1
+            if i3 < i2:
+                tmp_str += line[i3:i2]
+                ranges.append([i3, i2])
+            if i3 == len(line):
+                tmp_str += line[i3:i2]
+                ranges.append([i3, i2])
+            sections[ilev].append([ranges, tmp_str])
+    return sections
 
 
 def get_var_stack(line):
@@ -268,7 +280,7 @@ def apply_change(contents_split, change):
         if text[-1] == "\n" or text[-1] == "\r":
             text_split.append("")
 
-    if not change_range:
+    if change_range is None:
         # The whole file has changed
         return text_split, -1
 
@@ -279,7 +291,7 @@ def apply_change(contents_split, change):
 
     # Check for an edit occuring at the very end of the file
     if start_line == len(contents_split):
-        return contents_split.extend(text_split), -1
+        return contents_split + text_split, -1
 
     # Check for single line edit
     if start_line == end_line and len(text_split) == 1:
@@ -321,6 +333,7 @@ class LangServer:
         self.streaming = True
         self.symbol_include_mem = True
         self.autocomplete_no_prefix = False
+        self.lowercase_intrinsics = False
         self.sync_type = 1
         if logLevel == 0:
             logging.basicConfig(level=logging.ERROR)
@@ -333,6 +346,19 @@ class LangServer:
                 self.sync_type = settings["sync_type"]
             if "autocomplete_no_prefix" in settings:
                 self.autocomplete_no_prefix = settings["autocomplete_no_prefix"]
+            if "lowercase_intrinsics" in settings:
+                self.lowercase_intrinsics = settings["lowercase_intrinsics"]
+        if self.lowercase_intrinsics:
+            set_lowercase_intrinsics()
+        self.intrinsics = get_intrinsics()
+        for module in get_intrinsic_modules():
+            self.obj_tree[module.FQSN] = [module, None]
+
+    def post_message(self, message, type=1):
+        self.conn.send_notification("window/showMessage", {
+            "type": type,
+            "message": message
+        })
 
     def run(self):
         # Run server
@@ -347,10 +373,7 @@ class LangServer:
                 break
             else:
                 for message in self.post_messages:
-                    self.conn.send_notification("window/showMessage", {
-                        "type": message[0],
-                        "message": message[1]
-                    })
+                    self.post_message(message[1], message[0])
                 self.post_messages = []
 
     def handle(self, request):
@@ -442,12 +465,12 @@ class LangServer:
         #
         return {
             "capabilities": {
-                "documentSymbolProvider": True,
                 "completionProvider": {
                     "resolveProvider": False,
                     "triggerCharacters": ["%"]
                 },
                 "definitionProvider": True,
+                "documentSymbolProvider": True,
                 "referencesProvider": True,
                 "hoverProvider": True,
                 "textDocumentSync": self.sync_type
@@ -488,7 +511,7 @@ class LangServer:
         for scope in file_obj.get_scopes():
             scope_tree = scope.FQSN.split("::")
             if len(scope_tree) > 2:
-                if scope_tree[1].startswith("gen_int"):
+                if scope_tree[1].startswith("#gen_int"):
                     scope_type = 11
                 else:
                     continue
@@ -536,55 +559,54 @@ class LangServer:
             if type == 1:
                 return 9
             elif type == 4:
-                return 5
+                return 7
             elif type == 6:
                 return 6
             else:
                 return type
 
-        def get_candidates(scope_list, var_prefix, inc_globals=True, public_only=False):
+        def get_candidates(scope_list, var_prefix, inc_globals=True, public_only=False, abstract_only=False):
+            #
+            def child_candidates(scope, only_list=[], filter_public=True, req_abstract=False):
+                tmp_list = []
+                # Filter children
+                nonly = len(only_list)
+                for child in scope.get_children(filter_public):
+                    if req_abstract:
+                        if child.is_abstract():
+                            tmp_list += child_candidates(child, only_list, filter_public)
+                    else:
+                        if child.is_external_int():
+                            tmp_list += child_candidates(child, only_list, filter_public)
+                        else:
+                            if (nonly > 0) and (child.name not in only_list):
+                                continue
+                            tmp_list.append(child)
+                return tmp_list
             var_list = []
             use_dict = {}
             for scope in scope_list:
-                # Filter children
-                def_vis = scope.def_vis
-                for child in scope.get_children():
-                    if public_only:
-                        if child.vis < 0:
-                            continue
-                        if def_vis < 0 and child.vis <= 0:
-                            continue
-                    if child.name.lower().startswith(var_prefix):
-                        var_list.append(child)
-                # Add to use list
+                var_list += child_candidates(scope, filter_public=public_only, req_abstract=abstract_only)
+                # Traverse USE tree and add to list
                 use_dict = get_use_tree(scope, use_dict, self.obj_tree)
             # Look in found use modules
             for use_mod, only_list in use_dict.items():
                 scope = self.obj_tree[use_mod][0]
-                # Filter children
-                tmp_vars = []
-                def_vis = scope.def_vis
-                for child in scope.get_children():
-                    if child.vis < 0:
-                        continue
-                    if def_vis < 0 and child.vis <= 0:
-                        continue
-                    if child.name.lower().startswith(var_prefix):
-                        tmp_vars.append(child)
-                # Filter by ONLY statement
-                if len(only_list) > 0:
-                    for poss_var in tmp_vars:
-                        if poss_var.name in only_list:
-                            var_list.append(poss_var)
-                else:
-                    var_list.extend(tmp_vars)
+                var_list += child_candidates(scope, only_list, req_abstract=abstract_only)
             # Add globals
             if inc_globals:
-                for key in self.obj_tree:
-                    global_obj = self.obj_tree[key][0]
-                    if global_obj.name.lower().startswith(var_prefix):
-                        var_list.append(self.obj_tree[key][0])
-            return var_list
+                for key, obj in self.obj_tree.items():
+                    var_list.append(obj[0])
+                var_list += self.intrinsics
+            # Filter by prefix if necessary
+            if var_prefix == '':
+                return var_list
+            else:
+                tmp_list = []
+                for var in var_list:
+                    if var.name.lower().startswith(var_prefix):
+                        tmp_list.append(var)
+                return tmp_list
 
         def build_comp(candidate, name_only=False, name_replace=None):
             comp_obj = {}
@@ -597,32 +619,23 @@ class LangServer:
                     comp_obj["insertTextFormat"] = 2
             comp_obj["kind"] = map_types(candidate.get_type())
             comp_obj["detail"] = candidate.get_desc()
-            doc_str = candidate.get_documentation()
+            doc_str, _ = candidate.get_documentation()
             if doc_str is not None:
                 comp_obj["documentation"] = doc_str
             return comp_obj
 
         def get_context(line, var_prefix):
-            # tmp_prefix = var_prefix
             line_grouped = tokenize_line(line)
-            # Test if in call statement
             lev1_end = line_grouped[0][0][0][-1][1]
             if lev1_end < 0:
                 lev1_end = len(line)
-            if lev1_end == len(line):
-                test_match = CALL_REGEX.match(line_grouped[0][0][1])
-                if test_match is not None:
-                    return 3, var_prefix, None
-            # Test if variable definition using type/class
-            if len(line_grouped) >= 2:
-                lev2_end = line_grouped[1][0][0][-1][1]
-                if lev2_end < 0:
-                    lev2_end = len(line)
-                if lev2_end == len(line) and \
-                   line_grouped[1][0][0][-1][0] == lev1_end + 1:
-                    test_match = TYPE_STMNT_REGEX.match(line_grouped[0][0][1])
-                    if test_match is not None:
-                        return 4, var_prefix, None
+            # Test if variable definition statement
+            test_match = read_var_def(line)
+            if test_match is not None:
+                if test_match[0] == 'var':
+                    if (test_match[1][2] is None) and (lev1_end == len(line)):
+                        return 8, var_prefix, None
+                    return 7, var_prefix, None
             # Test if in USE statement
             test_match = read_use_stmt(line)
             if test_match is not None:
@@ -630,8 +643,38 @@ class LangServer:
                     return 2, var_prefix, test_match[1][0]
                 else:
                     return 1, var_prefix, None
+            # Test if scope declaration or end statement
+            if SCOPE_DEF_REGEX.match(line) or END_REGEX.match(line):
+                return -1, None, None
+            # Test if import statement
+            if IMPORT_REGEX.match(line):
+                return 5, var_prefix, None
+            # In type-def
+            type_def = False
+            if TYPE_DEF_REGEX.match(line) is not None:
+                type_def = True
+            # Test if in call statement
+            if lev1_end == len(line):
+                if CALL_REGEX.match(line_grouped[0][0][1]) is not None:
+                    return 3, var_prefix, None
+            # Test if variable definition using type/class or procedure
+            if len(line_grouped) >= 2:
+                lev2_end = line_grouped[1][0][0][-1][1]
+                if lev2_end < 0:
+                    lev2_end = len(line)
+                if (lev2_end == len(line)
+                        and line_grouped[1][0][0][-1][0] == lev1_end + 1):
+                    test_str = line_grouped[0][0][1]
+                    if ((TYPE_STMNT_REGEX.match(test_str) is not None)
+                            or (type_def and EXTENDS_REGEX.search(test_str) is not None)):
+                        return 4, var_prefix, None
+                    if PROCEDURE_STMNT_REGEX.match(test_str) is not None:
+                        return 6, var_prefix, None
             # Default context
-            return 0, var_prefix, None
+            if type_def:
+                return -1, var_prefix, None
+            else:
+                return 0, var_prefix, None
         # Get parameters from request
         req_dict = {"isIncomplete": False, "items": []}
         params = request["params"]
@@ -650,7 +693,8 @@ class LangServer:
         try:
             line_prefix = curr_line[:ac_char].lower()
             # Ignore for comment lines
-            if detect_comment_start(line_prefix, self.workspace[path]["fixed"]) >= 0:
+            comm_start = detect_comment_start(line_prefix, self.workspace[path]["fixed"])
+            if (comm_start >= 0) or (line_prefix[0] == '#'):
                 return req_dict
             # Ignore string literals
             if (line_prefix.count("'") % 2 == 1) or \
@@ -668,33 +712,14 @@ class LangServer:
         # Get context
         name_only = False
         public_only = False
+        include_globals = True
         line_context, var_prefix, context_info = \
             get_context(line_prefix, var_prefix)
-        if var_prefix == '' and not (is_member or line_context == 2):
+        if (line_context < 0) or (var_prefix == '' and not (is_member or line_context == 2)):
             return req_dict
         if self.autocomplete_no_prefix:
             var_prefix = ''
-        # USE stmnt
-        if line_context == 1:  # module part
-            for key in self.obj_tree:
-                candidate = self.obj_tree[key][0]
-                candidate_type = candidate.get_type()
-                if candidate_type != 1:
-                    continue
-                if candidate.name.lower().startswith(var_prefix):
-                    item_list.append(build_comp(candidate, name_only=True))
-            req_dict["items"] = item_list
-            return req_dict
-        elif line_context == 2:  # only part
-            name_only = True
-            mod_name = context_info.lower()
-            if mod_name in self.obj_tree:
-                scope_list = [self.obj_tree[mod_name][0]]
-                public_only = True
-            else:
-                return {"isIncomplete": False, "items": []}
-        #
-        include_globals = True
+        # Suggestions for user-defined type members
         if is_member:
             curr_scope = file_obj.get_inner_scope(ac_line+1)
             type_scope = climb_type_tree(var_stack, curr_scope, self.obj_tree)
@@ -704,22 +729,83 @@ class LangServer:
             else:
                 include_globals = False
                 scope_list = [type_scope]
-        #
-        for candidate in get_candidates(scope_list, var_prefix, include_globals, public_only):
+        # Setup based on context
+        req_callable = False
+        abstract_only = False
+        type_mask = [False for i in range(8)]
+        type_mask[1] = True
+        type_mask[4] = True
+        if line_context == 1:
+            # Use statement module part (modules only)
+            for key in self.obj_tree:
+                candidate = self.obj_tree[key][0]
+                candidate_type = candidate.get_type()
+                if candidate_type != 1:
+                    continue
+                if candidate.name.lower().startswith(var_prefix):
+                    item_list.append(build_comp(candidate, name_only=True))
+            req_dict["items"] = item_list
+            return req_dict
+        elif line_context == 2:
+            # Use statement only part (module public members only)
+            name_only = True
+            mod_name = context_info.lower()
+            if mod_name in self.obj_tree:
+                scope_list = [self.obj_tree[mod_name][0]]
+                public_only = True
+                include_globals = False
+                type_mask[4] = False
+            else:
+                return {"isIncomplete": False, "items": []}
+        elif line_context == 3:
+            # Filter callables for call statements
+            req_callable = True
+        elif line_context == 4:
+            # Variable definition statement for user-defined type
+            # (user-defined types only)
+            type_mask = [True for i in range(8)]
+            type_mask[4] = False
+        elif line_context == 5:
+            # Include statement (variables and user-defined types only)
+            name_only = True
+            type_mask = [True for i in range(8)]
+            type_mask[4] = False
+            type_mask[6] = False
+        elif line_context == 6:
+            # Variable definition statement for procedure with interface
+            # (interfaces only)
+            abstract_only = True
+            include_globals = False
+            name_only = True
+            type_mask = [True for i in range(8)]
+            type_mask[2] = False
+            type_mask[3] = False
+        elif line_context == 7:
+            # Variable definition statement (variables only)
+            name_only = True
+            type_mask[2] = True
+            type_mask[3] = True
+        elif line_context == 8:
+            # Variable definition keywords (variables only)
+            key_context = 0
+            enc_scope_type = scope_list[-1].get_type()
+            if enc_scope_type == 1:
+                key_context = 1
+            elif (enc_scope_type == 2) or (enc_scope_type == 3):
+                key_context = 2
+            elif enc_scope_type == 4:
+                key_context = 3
+            for candidate in get_keywords(key_context):
+                if candidate.name.lower().startswith(var_prefix):
+                    item_list.append(build_comp(candidate))
+            req_dict["items"] = item_list
+            return req_dict
+        for candidate in get_candidates(scope_list, var_prefix, include_globals, public_only, abstract_only):
             # Skip module names (only valid in USE)
             candidate_type = candidate.get_type()
-            if candidate_type == 1:
+            if type_mask[candidate_type]:
                 continue
-            # Only include types during variable definitions
-            # or select statements
-            if line_context == 4:
-                if candidate_type != 4:
-                    continue
-            elif line_context != 2:
-                if candidate_type == 4:
-                    continue
-            # Filter callables for call statements
-            if line_context == 3 and (not candidate.is_callable()):
+            if req_callable and (not candidate.is_callable()):
                 continue
             #
             if candidate_type == 5:
@@ -746,7 +832,8 @@ class LangServer:
         try:
             line_prefix = curr_line[:def_char].lower()
             # Ignore for comment lines
-            if detect_comment_start(line_prefix, def_file["fixed"]) >= 0:
+            comm_start = detect_comment_start(line_prefix, def_file["fixed"])
+            if (comm_start >= 0) or (line_prefix[0] == '#'):
                 return None
             # Ignore string literals
             if (line_prefix.count("'") % 2 == 1) or \
@@ -757,7 +844,7 @@ class LangServer:
             def_name = expand_name(curr_line, def_char)
         except:
             return None
-        # print(var_stack)
+        # print(var_stack, def_name)
         if def_name == '':
             return None
         file_obj = def_file["ast"]
@@ -779,6 +866,9 @@ class LangServer:
             key = def_name.lower()
             if key in self.obj_tree:
                 return self.obj_tree[key][0]
+            for obj in self.intrinsics:
+                if obj.name.lower() == key:
+                    return obj
         else:
             return var_obj
         return None
@@ -807,9 +897,11 @@ class LangServer:
             for filename, file_obj in sorted(self.workspace.items()):
                 # Search through file line by line
                 for (i, line) in enumerate(file_obj["contents"]):
+                    if len(line) == 0:
+                        continue
                     # Skip comment lines
                     comm_start = detect_comment_start(line, file_obj["fixed"])
-                    if comm_start == 0:
+                    if (comm_start == 0) or (line[0] == '#'):
                         continue
                     elif comm_start > 0:
                         line = line[:comm_start]
@@ -869,26 +961,19 @@ class LangServer:
             var_type = var_obj.get_type()
             # Currently only show for subroutines
             if var_type == 2 or var_type == 3:
-                skip_arg = False
-                hover_str, _ = var_obj.get_snippet()
-                hover_str += "\n"
-                if isinstance(var_obj, fortran_meth):
-                    if var_obj.modifiers.count(6) == 0:
-                        skip_arg = True
-                    var_obj = var_obj.link_obj
-                for arg_obj in var_obj.arg_objs:
-                    if skip_arg:
-                        skip_arg = False
-                        continue
-                    tmp_str = "  " + arg_obj.get_documentation()
-                    tmp_str += " :: {0}".format(arg_obj.name)
-                    hover_str = hover_str + tmp_str + "\n"
-                return {
-                    "contents": {
-                        "language": "fortran",
-                        "value": hover_str
-                    }
-                }
+                hover_str, highlight = var_obj.get_documentation(long=True)
+                if hover_str is not None:
+                    if highlight:
+                        return {
+                            "contents": {
+                                "language": "fortran",
+                                "value": hover_str
+                            }
+                        }
+                    else:
+                        return {
+                            "contents": hover_str
+                        }
         else:
             return None
 
@@ -918,10 +1003,9 @@ class LangServer:
         params = request["params"]
         uri = params["textDocument"]["uri"]
         path = path_from_uri(uri)
-        # Update file with changes
+        # Update file contents with changes
         if self.sync_type == 1:
             new_contents = params["contentChanges"][0]["text"].splitlines()
-            self.update_workspace_file(new_contents, path, update_links=True)
         else:
             if path in self.workspace:
                 new_contents = self.workspace[path]["contents"]
@@ -930,17 +1014,20 @@ class LangServer:
                         old_contents = new_contents
                         new_contents, line_tmp = apply_change(old_contents, change)
                 except:
-                    self.conn.send_notification("window/showMessage", {
-                        "type": 1,
-                        "message": 'Change request failed for unknown file "{0}"'.format(path)
-                    })
-                else:
-                    self.update_workspace_file(new_contents, path, update_links=True)
+                    self.post_message('Change request failed for file "{0}": Could not apply change'.format(path))
+                    return
             else:
-                self.conn.send_notification("window/showMessage", {
-                    "type": 1,
-                    "message": 'Change request failed for unknown file "{0}"'.format(path)
-                })
+                self.post_message('Change request failed for unknown file "{0}"'.format(path))
+                return
+        # Parse newly updated file
+        err_str = self.update_workspace_file(new_contents, path, update_links=True)
+        if err_str is not None:
+            self.post_message('Change request failed for file "{0}": {1}'.format(path, err_str))
+            return
+        # Update include statements linking to this file
+        for tmp_path, file_obj in self.workspace.items():
+            file_obj["ast"].resolve_includes(self.workspace, path=path)
+        self.workspace[path]["ast"].resolve_includes(self.workspace)
         # Update inheritance (currently only on open/save)
         # for key in self.obj_tree:
         #     self.obj_tree[key][0].resolve_inherit(self.obj_tree)
@@ -950,7 +1037,14 @@ class LangServer:
         # Update workspace from file on disk
         params = request["params"]
         uri = params["textDocument"]["uri"]
-        self.add_file(path_from_uri(uri))
+        filepath = path_from_uri(uri)
+        err_str = self.add_file(filepath)
+        if err_str is not None:
+            self.post_message('Save request failed for file "{0}": {1}'.format(filepath, err_str))
+            return
+        # Update include statements
+        for path, file_obj in self.workspace.items():
+            file_obj["ast"].resolve_includes(self.workspace)
         # Update inheritance
         for key in self.obj_tree:
             self.obj_tree[key][0].resolve_inherit(self.obj_tree)
@@ -959,12 +1053,10 @@ class LangServer:
 
     def add_file(self, filepath):
         # Read and add file from disk
-        if PY3K:
-            with open(filepath, 'r', encoding="utf-8") as fhandle:
-                self.update_workspace_file(fhandle.readlines(), filepath)
-        else:
-            with open(filepath, 'r') as fhandle:
-                self.update_workspace_file(fhandle.readlines(), filepath)
+        contents_split, err_str = read_file_split(filepath)
+        if contents_split is None:
+            return err_str
+        return self.update_workspace_file(contents_split, filepath)
 
     def update_workspace_file(self, contents_split, filepath, update_links=False):
         # Update workspace from file contents and path
@@ -972,32 +1064,28 @@ class LangServer:
             fixed_flag = detect_fixed_format(contents_split)
             ast_new = process_file(contents_split, True, filepath, fixed_flag)
         except:
-            self.conn.send_notification("window/showMessage", {
-                "type": 1,
-                "message": 'Parsing failed for file "{0}"'.format(filepath)
-            })
-            return  # Error during parsing
-        else:
-            # Remove old objects from tree
-            if filepath in self.workspace:
-                ast_old = self.workspace[filepath]["ast"]
-                for key in ast_old.global_dict:
-                    self.obj_tree.pop(key, None)
-            # Construct new file object and add to workspace
-            tmp_obj = {
-                "contents": contents_split,
-                "ast": ast_new,
-                "fixed": fixed_flag
-            }
-            self.workspace[filepath] = tmp_obj
-            # Add top-level objects to object tree
+            return 'Error during parsing'  # Error during parsing
+        # Remove old objects from tree
+        if filepath in self.workspace:
+            ast_old = self.workspace[filepath]["ast"]
+            for key in ast_old.global_dict:
+                self.obj_tree.pop(key, None)
+        # Construct new file object and add to workspace
+        tmp_obj = {
+            "contents": contents_split,
+            "ast": ast_new,
+            "fixed": fixed_flag
+        }
+        self.workspace[filepath] = tmp_obj
+        # Add top-level objects to object tree
+        for key, obj in ast_new.global_dict.items():
+            self.obj_tree[key] = [obj, filepath]
+        # Update local links/inheritance if necessary
+        if update_links:
             for key, obj in ast_new.global_dict.items():
-                self.obj_tree[key] = [obj, filepath]
-            # Update local links/inheritance if necessary
-            if update_links:
-                for key, obj in ast_new.global_dict.items():
-                    obj.resolve_inherit(self.obj_tree)
-                    obj.resolve_link(self.obj_tree)
+                obj.resolve_inherit(self.obj_tree)
+                obj.resolve_link(self.obj_tree)
+        return None
 
     def workspace_init(self):
         # Get filenames
@@ -1020,14 +1108,17 @@ class LangServer:
         pool.join()
         for path, result in results.items():
             result_obj = result.get()
-            if result_obj is None:
-                self.post_messages.append([1, 'Parsing failed for file "{0}"'.format(path)])
+            if result_obj[0] is None:
+                self.post_messages.append([1, 'Initialization failed for file "{0}": {1}'.format(path, result_obj[1])])
                 continue
-            self.workspace[path] = result_obj
+            self.workspace[path] = result_obj[0]
             # Add top-level objects to object tree
             ast_new = self.workspace[path]["ast"]
             for key in ast_new.global_dict:
                 self.obj_tree[key] = [ast_new.global_dict[key], path]
+        # Update include statements
+        for path, file_obj in self.workspace.items():
+            file_obj["ast"].resolve_includes(self.workspace)
         # Update inheritance
         for key in self.obj_tree:
             self.obj_tree[key][0].resolve_inherit(self.obj_tree)
