@@ -1,10 +1,11 @@
 import copy
 import re
+import os
 WORD_REGEX = re.compile(r'[a-z_][a-z0-9_]*', re.I)
-CLASS_VAR_REGEX = re.compile(r'(TYPE|CLASS)[ \t]*\(', re.I)
+CLASS_VAR_REGEX = re.compile(r'(TYPE|CLASS)[ ]*\(', re.I)
 
 
-def parse_keywords(keywords):
+def map_keywords(keywords):
     modifiers = []
     for key in keywords:
         key_lower = key.lower()
@@ -85,10 +86,12 @@ def get_use_tree(scope, use_dict, obj_tree, only_list=[]):
         if use_mod in obj_tree:
             if use_mod in use_dict:
                 old_len = len(use_dict[use_mod])
-                if old_len > 0:
+                if (old_len > 0) and (len(merged_use_list) > 0):
                     for only_name in use_stmnt[1]:
                         if use_dict[use_mod].count(only_name) == 0:
                             use_dict[use_mod].append(only_name)
+                else:
+                    use_dict[use_mod] = []
                 # Skip if we have already visited module with the same only list
                 if old_len == len(use_dict[use_mod]):
                     continue
@@ -99,11 +102,26 @@ def get_use_tree(scope, use_dict, obj_tree, only_list=[]):
 
 
 def find_in_scope(scope, var_name, obj_tree):
+    def check_scope(local_scope, var_name_lower, filter_public=False):
+        for child in local_scope.get_children():
+            if child.name.startswith("#GEN_INT"):
+                tmp_var, tmp_scope = check_scope(child, var_name_lower, filter_public)
+                if tmp_var is not None:
+                    return tmp_var, tmp_scope
+            if filter_public:
+                if child.vis < 0:
+                    continue
+                if local_scope.def_vis < 0 and child.vis <= 0:
+                    continue
+            if child.name.lower() == var_name_lower:
+                return child, local_scope
+        return None, None
+    #
     var_name_lower = var_name.lower()
     # Check local scope
-    for child in scope.get_children():
-        if child.name.lower() == var_name_lower:
-            return child, scope
+    tmp_var, tmp_scope = check_scope(scope, var_name_lower)
+    if tmp_var is not None:
+        return tmp_var, tmp_scope
     # Setup USE search
     use_dict = get_use_tree(scope, {}, obj_tree)
     # Look in found use modules
@@ -116,15 +134,9 @@ def find_in_scope(scope, var_name, obj_tree):
         if len(only_list) > 0:
             if var_name_lower not in only_list:
                 continue
-        # Check for variable in public children
-        def_vis = use_scope.def_vis
-        for child in use_scope.get_children():
-            if child.vis < 0:
-                continue
-            if def_vis < 0 and child.vis <= 0:
-                continue
-            if child.name.lower() == var_name_lower:
-                return child, use_scope
+        tmp_var, tmp_scope = check_scope(use_scope, var_name_lower, filter_public=True)
+        if tmp_var is not None:
+            return tmp_var, tmp_scope
     # Check parent scopes
     if scope.parent is not None:
         curr_scope = scope.parent
@@ -192,11 +204,20 @@ class fortran_scope:
         for child in self.children:
             child.resolve_link(obj_tree)
 
-    def add_parent(self, parent_obj):
+    def set_parent(self, parent_obj):
         self.parent = parent_obj
 
     def add_child(self, child):
         self.children.append(child)
+        child.set_parent(self)
+
+    def update_fqsn(self, enc_scope=None):
+        if enc_scope is not None:
+            self.FQSN = enc_scope.lower() + "::" + self.name.lower()
+        else:
+            self.FQSN = self.name.lower()
+        for child in self.children:
+            child.update_fqsn(self.FQSN)
 
     def add_member(self, member):
         self.members.append(member)
@@ -212,11 +233,24 @@ class fortran_scope:
             return name_replace, None
         return self.name, None
 
-    def get_documentation(self):
-        return None
+    def get_documentation(self, long=False):
+        return None, False
 
-    def get_children(self):
-        return self.children
+    def get_children(self, public_only=False):
+        if public_only:
+            pub_children = []
+            for child in self.children:
+                if child.name.startswith("#GEN_INT"):
+                    pub_children.append(child)
+                    continue
+                if child.vis < 0:
+                    continue
+                if (self.def_vis < 0) and (child.vis <= 0):
+                    continue
+                pub_children.append(child)
+            return pub_children
+        else:
+            return self.children
 
     def get_ancestors(self):
         return []
@@ -230,6 +264,12 @@ class fortran_scope:
     def is_callable(self):
         return False
 
+    def is_external_int(self):
+        return False
+
+    def is_abstract(self):
+        return False
+
     def end(self, line_number):
         self.eline = line_number
 
@@ -238,16 +278,16 @@ class fortran_scope:
         if self.parent is not None:
             if self.parent.get_type() == 5:
                 return []
-        FSQN_list = []
+        FQSN_list = []
         errors = []
         for child in self.children:
             # Check other variables in current scope
-            if FSQN_list.count(child.FQSN) > 0:
+            if FQSN_list.count(child.FQSN) > 0:
                 line_number = child.sline - 1
                 i0, i1 = find_word_in_line(file_contents[line_number].lower(), child.name.lower())
                 errors.append([0, line_number, i0, i1, child.name])
             else:
-                FSQN_list.append(child.FQSN)
+                FQSN_list.append(child.FQSN)
             # Check for masking from parent scope in subroutines and functions
             if self.parent is not None and (self.get_type() == 2 or self.get_type() == 3):
                 parent_var, parent_scope = \
@@ -262,14 +302,10 @@ class fortran_scope:
         return errors
 
     def check_use(self, obj_tree, file_contents):
-        intrinsic_mods = ['omp_lib', 'iso_c_binding', 'iso_fortran_env',
-                          'ieee_exceptions', 'ieee_arithmetic', 'ieee_features']
         errors = []
         for use_line in self.use:
             use_mod = use_line[0]
             if use_mod not in obj_tree:
-                if use_mod in intrinsic_mods:
-                    continue
                 line_number = use_line[2] - 1
                 line = file_contents[line_number]
                 i0 = line.lower().find(use_mod)
@@ -337,7 +373,7 @@ class fortran_submodule(fortran_module):
 
 
 class fortran_subroutine(fortran_scope):
-    def __init__(self, file_obj, line_number, name, enc_scope=None, args=None, mod_sub=False):
+    def __init__(self, file_obj, line_number, name, enc_scope=None, args="", mod_sub=False):
         self.base_setup(file_obj, line_number, name, enc_scope)
         self.args = args
         self.arg_objs = []
@@ -362,10 +398,9 @@ class fortran_subroutine(fortran_scope):
         self.in_children = []
         for child in copy_source.arg_objs:
             if child.name.lower() not in child_names:
-                print(self.FSQN, child.FQSN)
                 self.in_children.append(child)
 
-    def get_children(self):
+    def get_children(self, public_only=False):
         tmp_list = copy.copy(self.children)
         tmp_list.extend(self.in_children)
         return tmp_list
@@ -423,9 +458,26 @@ class fortran_subroutine(fortran_scope):
     def get_desc(self):
         return 'SUBROUTINE'
 
+    def get_documentation(self, long=False):
+        if long:
+            skip_arg = False
+            hover_str, _ = self.get_snippet()
+            hover_str += "\n"
+            for arg_obj in self.arg_objs:
+                if skip_arg:
+                    skip_arg = False
+                    continue
+                arg_doc, _ = arg_obj.get_documentation()
+                tmp_str = "  " + arg_doc
+                tmp_str += " :: {0}".format(arg_obj.name)
+                hover_str = hover_str + tmp_str + "\n"
+            return hover_str, True
+        else:
+            return None, False
+
 
 class fortran_function(fortran_subroutine):
-    def __init__(self, file_obj, line_number, name, enc_scope=None, args=None,
+    def __init__(self, file_obj, line_number, name, enc_scope=None, args="",
                  mod_fun=False, return_type=None, result_var=None):
         self.base_setup(file_obj, line_number, name, enc_scope)
         self.args = args
@@ -436,7 +488,7 @@ class fortran_function(fortran_subroutine):
         self.result_obj = None
         if return_type is not None:
             self.return_type = return_type[0]
-            self.modifiers = parse_keywords(return_type[1])
+            self.modifiers = map_keywords(return_type[1])
         else:
             self.return_type = None
             self.modifiers = []
@@ -502,7 +554,7 @@ class fortran_type(fortran_scope):
     def get_desc(self):
         return 'TYPE'
 
-    def get_children(self):
+    def get_children(self, public_only=False):
         tmp_list = copy.copy(self.children)
         tmp_list.extend(self.in_children)
         return tmp_list
@@ -528,9 +580,11 @@ class fortran_type(fortran_scope):
 
 
 class fortran_int(fortran_scope):
-    def __init__(self, file_obj, line_number, name, enc_scope=None):
+    def __init__(self, file_obj, line_number, name, enc_scope=None, abstract=False):
         self.base_setup(file_obj, line_number, name, enc_scope)
         self.mems = []
+        self.abstract = abstract
+        self.external = name.startswith('#GEN_INT') and (not abstract)
 
     def get_type(self):
         return 5
@@ -541,6 +595,12 @@ class fortran_int(fortran_scope):
     def is_callable(self):
         return True
 
+    def is_external_int(self):
+        return self.external
+
+    def is_abstract(self):
+        return self.abstract
+
     def resolve_link(self, obj_tree):
         if self.parent is None:
             return
@@ -549,6 +609,8 @@ class fortran_int(fortran_scope):
             mem_obj, _ = find_in_scope(self.parent, member, obj_tree)
             if mem_obj is not None:
                 self.mems.append(mem_obj)
+        for child in self.children:
+            child.resolve_link(obj_tree)
 
 
 class fortran_obj:
@@ -578,9 +640,17 @@ class fortran_obj:
             elif modifier == 5:
                 self.vis = -1
 
-    def add_parent(self, parent_obj):
+    def set_parent(self, parent_obj):
         self.parent = parent_obj
 
+    def update_fqsn(self, enc_scope=None):
+        if enc_scope is not None:
+            self.FQSN = enc_scope.lower() + "::" + self.name.lower()
+        else:
+            self.FQSN = self.name.lower()
+        for child in self.children:
+            child.update_fqsn(self.FQSN)
+
     def resolve_link(self, obj_tree):
         if self.link_name is None:
             return
@@ -624,7 +694,7 @@ class fortran_obj:
         # Normal variable
         return name, None
 
-    def get_documentation(self):
+    def get_documentation(self, long=False):
         if self.link_obj is not None:
             return self.link_obj.get_documentation()
         #
@@ -632,9 +702,9 @@ class fortran_obj:
         if len(self.modifiers) > 0:
             doc_str += ", "
             doc_str += ", ".join(get_keywords(self.modifiers))
-        return doc_str
+        return doc_str, True
 
-    def get_children(self):
+    def get_children(self, public_only=False):
         return []
 
     def resolve_inherit(self, obj_tree):
@@ -649,6 +719,12 @@ class fortran_obj:
     def is_callable(self):
         return self.callable
 
+    def is_external_int(self):
+        return False
+
+    def is_abstract(self):
+        return False
+
 
 class fortran_meth(fortran_obj):
     def get_snippet(self, name_replace=None, drop_arg=None):
@@ -670,6 +746,33 @@ class fortran_meth(fortran_obj):
         # Generic
         return 7
 
+    def get_documentation(self, long=False):
+        if long:
+            skip_arg = False
+            hover_str, _ = self.get_snippet()
+            hover_str += "\n"
+            if self.modifiers.count(6) == 0:
+                skip_arg = True
+            var_obj = self.link_obj
+            for arg_obj in var_obj.arg_objs:
+                if skip_arg:
+                    skip_arg = False
+                    continue
+                arg_doc, _ = arg_obj.get_documentation()
+                tmp_str = "  " + arg_doc
+                tmp_str += " :: {0}".format(arg_obj.name)
+                hover_str = hover_str + tmp_str + "\n"
+            return hover_str, True
+        else:
+            if self.link_obj is not None:
+                return self.link_obj.get_documentation()
+            #
+            doc_str = self.desc
+            if len(self.modifiers) > 0:
+                doc_str += ", "
+                doc_str += ", ".join(get_keywords(self.modifiers))
+            return doc_str, True
+
     def is_callable(self):
         return True
 
@@ -685,6 +788,7 @@ class fortran_file:
         self.scope_stack = []
         self.end_stack = []
         self.pp_if = []
+        self.include_stmnts = []
         self.none_scope = None
         self.current_scope = None
         self.END_REGEX = None
@@ -694,7 +798,7 @@ class fortran_file:
         if self.none_scope is not None:
             raise ValueError
         self.none_scope = fortran_program(self, 1, "main")
-        self.add_scope(self.none_scope, re.compile(r'[ \t]*END[ \t]*PROGRAM', re.I), exportable=False)
+        self.add_scope(self.none_scope, re.compile(r'[ ]*END[ ]*PROGRAM', re.I), exportable=False)
 
     def get_enc_scope_name(self):
         if self.current_scope is None:
@@ -711,14 +815,12 @@ class fortran_file:
                 self.create_none_scope()
                 new_scope.FQSN = self.none_scope.FQSN + "::" + new_scope.name.lower()
                 self.current_scope.add_child(new_scope)
-                new_scope.add_parent(self.current_scope)
                 self.scope_stack.append(self.current_scope)
             else:
                 if exportable:
                     self.global_dict[new_scope.FQSN] = new_scope
         else:
             self.current_scope.add_child(new_scope)
-            new_scope.add_parent(self.current_scope)
             self.scope_stack.append(self.current_scope)
         if self.END_REGEX is not None:
             self.end_stack.append(self.END_REGEX)
@@ -743,7 +845,6 @@ class fortran_file:
             self.create_none_scope()
             new_var.FQSN = self.none_scope.FQSN + "::" + new_var.name.lower()
         self.current_scope.add_child(new_var)
-        new_var.add_parent(self.current_scope)
         self.variable_list.append(new_var)
 
     def add_int_member(self, key):
@@ -760,6 +861,11 @@ class fortran_file:
             self.create_none_scope()
         self.current_scope.add_use(mod_word, line_number, only_list)
 
+    def add_include(self, path, line_number):
+        if os.name == "nt":
+            path = path.replace("/", "\\")
+        self.include_stmnts.append([line_number, path, []])
+
     def start_ppif(self, line_number):
         self.pp_if.append([line_number-1, -1])
 
@@ -782,6 +888,8 @@ class fortran_file:
                 scope_list.append(scope)
                 for ancestor in scope.get_ancestors():
                     scope_list.append(ancestor)
+        if len(scope_list) == 0 and self.none_scope is not None:
+            return [self.none_scope]
         return scope_list
 
     def get_inner_scope(self, line_number):
@@ -792,6 +900,8 @@ class fortran_file:
                 if line_number >= scope.sline and line_number <= scope.eline:
                     curr_scope = scope
                     scope_sline = scope.sline
+        if curr_scope is None and self.none_scope is not None:
+            return self.none_scope
         return curr_scope
 
     def get_object(self, FQSN):
@@ -809,6 +919,13 @@ class fortran_file:
             for name in FQSN_split[1:]:
                 next_obj = None
                 for child in curr_obj.children:
+                    if child.name.startswith("#GEN_INT"):
+                        for int_child in child.get_children():
+                            if int_child.name == name:
+                                next_obj = int_child
+                                break
+                        if next_obj is not None:
+                            break
                     if child.name == name:
                         next_obj = child
                         break
@@ -817,7 +934,30 @@ class fortran_file:
                 curr_obj = next_obj
         return curr_obj
 
-    def close_file(self):
+    def resolve_includes(self, workspace, path=None):
+        file_dir = os.path.dirname(self.path)
+        for include_path in self.include_stmnts:
+            file_path = os.path.join(file_dir, include_path[1])
+            if path is not None:
+                if not (path == file_path):
+                    continue
+            parent_scope = self.get_inner_scope(include_path[0])
+            added_entities = include_path[2]
+            if file_path in workspace:
+                include_obj = workspace[file_path]["ast"]
+                if (include_obj.none_scope is not None) and (include_obj.none_scope is not parent_scope):
+                    # Remove old objects
+                    for obj in added_entities:
+                        parent_scope.children.remove(obj)
+                    added_entities = []
+                    for child in include_obj.none_scope.children:
+                        added_entities.append(child)
+                        parent_scope.add_child(child)
+                        child.update_fqsn(parent_scope.FQSN)
+                    include_obj.none_scope = parent_scope
+                    include_path[2] = added_entities
+
+    def close_file(self, line_number):
         # Tasks to be done when file parsing is finished
         for private_name in self.private_list:
             obj = self.get_object(private_name)
@@ -827,10 +967,16 @@ class fortran_file:
             obj = self.get_object(public_name)
             if obj is not None:
                 obj.set_visibility(1)
+        if self.none_scope is not None:
+            self.none_scope.end(line_number)
+            self.scope_list.remove(self.none_scope)
 
     def check_file(self, obj_tree, file_contents):
         errors = []
-        for scope in self.scope_list:
+        tmp_list = self.scope_list
+        if self.none_scope is not None:
+            tmp_list += [self.none_scope]
+        for scope in tmp_list:
             for error in scope.check_double_def(file_contents, obj_tree):
                 # Check preproc if
                 if self.check_ppif(error[1]):
