@@ -1,35 +1,36 @@
 from __future__ import print_function
 import re
-from fortls.objects import parse_keywords, fortran_module, fortran_program, \
+from fortls.objects import map_keywords, fortran_module, fortran_program, \
     fortran_submodule, fortran_subroutine, fortran_function, fortran_type, \
     fortran_int, fortran_obj, fortran_meth, fortran_file
 #
-USE_REGEX = re.compile(r'[ \t]*USE([, \t]*INTRINSIC)?[ \t:]*([a-z0-9_]*)', re.I)
-SUB_REGEX = re.compile(r'[ \t]*(PURE|ELEMENTAL|RECURSIVE)*[ \t]*(SUBROUTINE)', re.I)
-END_SUB_REGEX = re.compile(r'[ \t]*END[ \t]*SUBROUTINE', re.I)
-FUN_REGEX = re.compile(r'[ \t]*(PURE|ELEMENTAL|RECURSIVE)*[ \t]*(FUNCTION)', re.I)
+USE_REGEX = re.compile(r'[ ]*USE([, ]*INTRINSIC)?[ :]*([a-z0-9_]*)', re.I)
+INCLUDE_REGEX = re.compile(r'[ ]*INCLUDE[ :]*[\'\"]([^\'\"]*)', re.I)
+SUB_REGEX = re.compile(r'[ ]*(PURE|ELEMENTAL|RECURSIVE)*[ ]*SUBROUTINE[ ]+([a-z0-9_]+)', re.I)
+END_SUB_REGEX = re.compile(r'[ ]*END[ ]*SUBROUTINE', re.I)
+FUN_REGEX = re.compile(r'[ ]*(PURE|ELEMENTAL|RECURSIVE)*[ ]*FUNCTION[ ]+([a-z0-9_]+)', re.I)
 RESULT_REGEX = re.compile(r'RESULT[ ]*\(([a-z0-9_]*)\)', re.I)
-END_FUN_REGEX = re.compile(r'[ \t]*END[ \t]*FUNCTION', re.I)
-MOD_REGEX = re.compile(r'[ \t]*MODULE[ \t]*([a-z0-9_]*)', re.I)
-END_MOD_REGEX = re.compile(r'[ \t]*END[ \t]*MODULE', re.I)
-SUBMOD_REGEX = re.compile(r'[ \t]*SUBMODULE[ \t]*\(', re.I)
-END_SMOD_REGEX = re.compile(r'[ \t]*END[ \t]*SUBMODULE', re.I)
-PROG_REGEX = re.compile(r'[ \t]*PROGRAM[ \t]*([a-z0-9_]*)', re.I)
-END_PROG_REGEX = re.compile(r'[ \t]*END[ \t]*PROGRAM', re.I)
-INT_REGEX = re.compile(r'[ \t]*(?:ABSTRACT)?[ \t]*INTERFACE[ \t]*([a-z0-9_]*)', re.I)
-END_INT_REGEX = re.compile(r'[ \t]*END[ \t]*INTERFACE', re.I)
-END_GEN_REGEX = re.compile(r'[ \t]*END[ \t]*$', re.I)
-TYPE_DEF_REGEX = re.compile(r'[ \t]*TYPE', re.I)
+END_FUN_REGEX = re.compile(r'[ ]*END[ ]*FUNCTION', re.I)
+MOD_REGEX = re.compile(r'[ ]*MODULE[ ]+([a-z0-9_]+)', re.I)
+END_MOD_REGEX = re.compile(r'[ ]*END[ ]*MODULE', re.I)
+SUBMOD_REGEX = re.compile(r'[ ]*SUBMODULE[ ]*\(', re.I)
+END_SMOD_REGEX = re.compile(r'[ ]*END[ ]*SUBMODULE', re.I)
+PROG_REGEX = re.compile(r'[ ]*PROGRAM[ ]+([a-z0-9_]+)', re.I)
+END_PROG_REGEX = re.compile(r'[ ]*END[ ]*PROGRAM', re.I)
+INT_REGEX = re.compile(r'[ ]*(ABSTRACT)?[ ]*INTERFACE[ ]*([a-z0-9_]*)', re.I)
+END_INT_REGEX = re.compile(r'[ ]*END[ ]*INTERFACE', re.I)
+END_GEN_REGEX = re.compile(r'[ ]*END[ ]*$', re.I)
+TYPE_DEF_REGEX = re.compile(r'[ ]*(TYPE)[, ]+', re.I)
 EXTENDS_REGEX = re.compile(r'EXTENDS[ ]*\(([a-z0-9_]*)\)', re.I)
-END_TYPED_REGEX = re.compile(r'[ \t]*END[ \t]*TYPE', re.I)
-NAT_VAR_REGEX = re.compile(r'[ \t]*(INTEGER|REAL|DOUBLE PRECISION|COMPLEX'
+END_TYPED_REGEX = re.compile(r'[ ]*END[ ]*TYPE', re.I)
+NAT_VAR_REGEX = re.compile(r'[ ]*(INTEGER|REAL|DOUBLE PRECISION|COMPLEX'
                            r'|CHARACTER|LOGICAL|PROCEDURE|CLASS|TYPE)', re.I)
-KIND_SPEC_REGEX = re.compile(r'([ \t]*\([a-z0-9_ =*]*\)|\*[0-9]*)', re.I)
-KEYWORD_LIST_REGEX = re.compile(r'[ \t]*,[ \t]*(PUBLIC|PRIVATE|ALLOCATABLE|'
+KIND_SPEC_REGEX = re.compile(r'([ ]*\([a-z0-9_ =*]*\)|\*[0-9]*)', re.I)
+KEYWORD_LIST_REGEX = re.compile(r'[ ]*,[ ]*(PUBLIC|PRIVATE|ALLOCATABLE|'
                                 r'POINTER|TARGET|DIMENSION\([a-z0-9_:, ]*\)|'
                                 r'OPTIONAL|INTENT\([inout]*\)|DEFERRED|NOPASS|'
                                 r'SAVE|PARAMETER)', re.I)
-TATTR_LIST_REGEX = re.compile(r'[ \t]*,[ \t]*(PUBLIC|PRIVATE|ABSTRACT|EXTENDS\([a-z0-9_]*\))', re.I)
+TATTR_LIST_REGEX = re.compile(r'[ ]*,[ ]*(PUBLIC|PRIVATE|ABSTRACT|EXTENDS\([a-z0-9_]*\))', re.I)
 VIS_REGEX = re.compile(r'(PUBLIC|PRIVATE)', re.I)
 WORD_REGEX = re.compile(r'[a-z_][a-z0-9_]*', re.I)
 SUB_PAREN_MATCH = re.compile(r'\([a-z0-9_, ]*\)', re.I)
@@ -40,10 +41,10 @@ DQ_STRING_REGEX = re.compile(r'\"[^\"]*\"', re.I)
 FIXED_COMMENT_LINE_MATCH = re.compile(r'(!|c|d|\*)', re.I)
 FIXED_CONT_REGEX = re.compile(r'(     [\S])')
 #
-FREE_COMMENT_LINE_MATCH = re.compile(r'([ \t]*!)')
-FREE_CONT_REGEX = re.compile(r'([ \t]*&)')
+FREE_COMMENT_LINE_MATCH = re.compile(r'([ ]*!)')
+FREE_CONT_REGEX = re.compile(r'([ ]*&)')
 FREE_FORMAT_TEST = re.compile(r'[ ]{1,4}[a-z]', re.I)
-OPENMP_LINE_MATCH = re.compile(r'[ \t]*[!|c|\*]\$OMP', re.I)
+OPENMP_LINE_MATCH = re.compile(r'[ ]*[!|c|\*]\$OMP', re.I)
 #
 PPIF_START_TEST = re.compile(r'#if')
 PPIF_END_TEST = re.compile(r'#endif')
@@ -94,16 +95,30 @@ def separate_def_list(test_str):
         elif char == ')':
             paren_count -= 1
         elif char == ',' and paren_count == 0:
+            curr_str = curr_str.strip()
             if curr_str != '':
                 def_list.append(curr_str)
                 curr_str = ''
+            elif (curr_str == '') and (len(def_list) == 0):
+                return None
             continue
         curr_str += char
+    curr_str = curr_str.strip()
     if curr_str != '':
         def_list.append(curr_str)
     return def_list
 
 
+def parse_keywords(test_str):
+    keyword_match = KEYWORD_LIST_REGEX.match(test_str)
+    keywords = []
+    while (keyword_match is not None):
+        keywords.append(keyword_match.group(0).replace(',', ' ').strip().upper())
+        test_str = test_str[keyword_match.end(0):]
+        keyword_match = KEYWORD_LIST_REGEX.match(test_str)
+    return keywords, test_str
+
+
 def get_var_dims(test_str):
     paren_count = 0
     curr_dim = 0
@@ -146,12 +161,7 @@ def read_var_def(line, type_word=None):
         if trailing_line[0] != ' ' and trailing_line[0] != ',':
             return None
     #
-    keyword_match = KEYWORD_LIST_REGEX.match(trailing_line)
-    keywords = []
-    while (keyword_match is not None):
-        keywords.append(keyword_match.group(0).replace(',', ' ').strip().upper())
-        trailing_line = trailing_line[keyword_match.end(0):]
-        keyword_match = KEYWORD_LIST_REGEX.match(trailing_line)
+    keywords, trailing_line = parse_keywords(trailing_line)
     # Check if function
     fun_def = read_fun_def(trailing_line, [type_word, keywords])
     if fun_def is not None:
@@ -160,13 +170,15 @@ def read_var_def(line, type_word=None):
     line_split = trailing_line.split('::')
     if len(line_split) == 1:
         if len(keywords) > 0:
-            return None
+            var_words = None
         else:
             trailing_line = line_split[0]
+            var_words = separate_def_list(trailing_line.strip())
     else:
         trailing_line = line_split[1]
-    #
-    var_words = separate_def_list(trailing_line.strip())
+        var_words = separate_def_list(trailing_line.strip())
+        if var_words is None:
+            var_words = []
     #
     return 'var', [type_word, keywords, var_words]
 
@@ -176,14 +188,9 @@ def read_fun_def(line, return_type=None, mod_fun=False):
     if fun_match is None:
         return None
     #
-    trailing_line = line[fun_match.end(0):].strip()
-    trailing_line = trailing_line.split('!')[0]
-    name_match = WORD_REGEX.match(trailing_line)
-    if name_match is not None:
-        name = name_match.group(0)
-        trailing_line = trailing_line[name_match.end(0):].strip()
-    else:
-        return None
+    name = fun_match.group(2)
+    trailing_line = line[fun_match.end(0):].split('!')[0]
+    trailing_line = trailing_line.strip()
     #
     paren_match = SUB_PAREN_MATCH.match(trailing_line)
     args = ''
@@ -208,14 +215,9 @@ def read_sub_def(line, mod_sub=False):
     if sub_match is None:
         return None
     #
-    trailing_line = line[sub_match.end(0):].strip()
-    trailing_line = trailing_line.split('!')[0]
-    name_match = WORD_REGEX.match(trailing_line)
-    if name_match is not None:
-        name = name_match.group(0)
-        trailing_line = trailing_line[name_match.end(0):].strip()
-    else:
-        return None
+    name = sub_match.group(2)
+    trailing_line = line[sub_match.end(0):].split('!')[0]
+    trailing_line = trailing_line.strip()
     #
     paren_match = SUB_PAREN_MATCH.match(trailing_line)
     args = ''
@@ -232,8 +234,8 @@ def read_type_def(line):
     type_match = TYPE_DEF_REGEX.match(line)
     if type_match is None:
         return None
-    trailing_line = line[type_match.end(0):]
-    trailing_line = trailing_line.split('!')[0]
+    trailing_line = line[type_match.end(1):].split('!')[0]
+    trailing_line = trailing_line.strip()
     # Parse keywords
     keyword_match = TATTR_LIST_REGEX.match(trailing_line)
     keywords = []
@@ -300,8 +302,8 @@ def read_submod_def(line):
     else:
         parent_name = None
         name = None
-        trailing_line = line[submod_match.end(0):].strip()
-        trailing_line = trailing_line.split('!')[0]
+        trailing_line = line[submod_match.end(0):].split('!')[0]
+        trailing_line = trailing_line.strip()
         parent_match = WORD_REGEX.match(trailing_line)
         if parent_match is not None:
             parent_name = parent_match.group(0).lower()
@@ -329,12 +331,13 @@ def read_int_def(line):
     if int_match is None:
         return None
     else:
-        int_name = int_match.group(1).lower()
+        int_name = int_match.group(2).lower()
+        is_abstract = int_match.group(1) is not None
         if int_name == '':
-            return 'int', None
+            return 'int', [None, is_abstract]
         if int_name == 'assignment' or int_name == 'operator':
             return None
-        return 'int', int_match.group(1)
+        return 'int', [int_match.group(2), is_abstract]
 
 
 def read_use_stmt(line):
@@ -353,8 +356,18 @@ def read_use_stmt(line):
         return 'use', [use_mod, only_list]
 
 
+def read_inc_stmt(line):
+    inc_match = INCLUDE_REGEX.match(line)
+    if inc_match is None:
+        return None
+    else:
+        inc_path = inc_match.group(1)
+        return 'inc', [inc_path]
+
+
 def_tests = [read_var_def, read_sub_def, read_fun_def, read_type_def,
-             read_use_stmt, read_int_def, read_mod_def, read_prog_def, read_submod_def]
+             read_use_stmt, read_int_def, read_mod_def, read_prog_def,
+             read_submod_def, read_inc_stmt]
 
 
 def process_file(file_str, close_open_scopes, path=None, fixed_format=False, debug=False):
@@ -373,7 +386,7 @@ def process_file(file_str, close_open_scopes, path=None, fixed_format=False, deb
     # at_eof = False
     next_line = None
     line_ind = 0
-    while(line_ind < len(file_str)):
+    while((line_ind < len(file_str)) or (next_line is not None)):
         # Get next line
         if next_line is None:
             line = file_str[line_ind]
@@ -465,6 +478,8 @@ def process_file(file_str, close_open_scopes, path=None, fixed_format=False, deb
             obj_type = obj_read[0]
             obj = obj_read[1]
             if obj_type == 'var':
+                if obj[2] is None:
+                    continue
                 link_name = None
                 if obj[0][:3] == 'PRO':
                     if isinstance(file_obj.current_scope, fortran_int):
@@ -490,7 +505,7 @@ def process_file(file_str, close_open_scopes, path=None, fixed_format=False, deb
                     if name_stripped.find('(') > -1:
                         var_dim = get_var_dims(name_stripped)
                     name_stripped = name_stripped.split('(')[0].strip()
-                    modifiers = parse_keywords(obj[1])
+                    modifiers = map_keywords(obj[1])
                     if obj[0][:3] == 'PRO':
                         new_var = fortran_meth(file_obj, line_number, name_stripped, obj[0],
                                                modifiers, file_obj.enc_scope_name, link_name)
@@ -533,7 +548,7 @@ def process_file(file_str, close_open_scopes, path=None, fixed_format=False, deb
                 if(debug):
                     print('{1} !!! FUNCTION statement({0})'.format(line_number, line.strip()))
             elif obj_type == 'typ':
-                modifiers = parse_keywords(obj[2])
+                modifiers = map_keywords(obj[2])
                 new_type = fortran_type(file_obj, line_number, obj[0], modifiers, file_obj.enc_scope_name)
                 if obj[1] is not None:
                     new_type.set_inherit(obj[1])
@@ -542,11 +557,13 @@ def process_file(file_str, close_open_scopes, path=None, fixed_format=False, deb
                     print('{1} !!! TYPE statement({0})'.format(line_number, line.strip()))
             elif obj_type == 'int':
                 hidden = False
-                if obj is None:
+                abstract = obj[1]
+                name = obj[0]
+                if name is None:
                     int_counter += 1
-                    obj = 'GEN_INT{0}'.format(int_counter)
+                    name = '#GEN_INT{0}'.format(int_counter)
                     hidden = True
-                new_int = fortran_int(file_obj, line_number, obj, file_obj.enc_scope_name)
+                new_int = fortran_int(file_obj, line_number, name, file_obj.enc_scope_name, abstract)
                 file_obj.add_scope(new_int, END_INT_REGEX, hidden=hidden, req_container=True)
                 if(debug):
                     print('{1} !!! INTERFACE statement({0})'.format(line_number, line.strip()))
@@ -563,6 +580,10 @@ def process_file(file_str, close_open_scopes, path=None, fixed_format=False, deb
                 file_obj.add_use(obj[0], line_number, obj[1])
                 if(debug):
                     print('{1} !!! USE statement({0})'.format(line_number, line.strip()))
+            elif obj_type == 'inc':
+                file_obj.add_include(obj[0], line_number)
+                if(debug):
+                    print('{1} !!! INCLUDE statement({0})'.format(line_number, line.strip()))
         # Look for visiblity statement
         match = VIS_REGEX.match(line)
         if (match is not None):
@@ -582,5 +603,5 @@ def process_file(file_str, close_open_scopes, path=None, fixed_format=False, deb
             if(debug):
                 print('{1} !!! Visiblity statement({0})'.format(line_number, line.strip()))
             continue
-    file_obj.close_file()
+    file_obj.close_file(line_number)
     return file_obj
